/*
 * Copyright (c) 2007 Kathryn Huxtable
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: AttributeFormatter.java 1223 2007-04-17 02:50:37Z kathryn5 $
 */
package org.kathrynhuxtable.middleware.shibshim.attributes;

import java.util.Date;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.kathrynhuxtable.middleware.shibshim.util.TimeUtil;

/**
 * Global constants needed to assemble the XML document.
 */
public final class AttributeFormatter {
    private static Logger       log               = Logger.getLogger(AttributeFormatter.class.getName());

    private static final String XML_STMT_TAG      = "AttributeStatement";
    private static final String XML_ATTR_TAG      = "Attribute";
    private static final String XML_ATTRNAME_ATTR = "AttributeName";
    private static final String XML_ATTRVAL_TAG   = "AttributeValue";

    private static final String XML_STMT_B_TAG    = "<" + XML_STMT_TAG + ">";
    private static final String XML_STMT_E_TAG    = "</" + XML_STMT_TAG + ">";
    private static final String XML_ATTR_E_TAG    = "</" + XML_ATTR_TAG + ">";
    private static final String XML_ATTRVAL_B_TAG = "<" + XML_ATTRVAL_TAG + ">";
    private static final String XML_ATTRVAL_E_TAG = "</" + XML_ATTRVAL_TAG + ">";

    /**
     * Assemble the attributes into an XML &lt;Assertion&gt; document.
     * 
     * @param attributes
     *            attribute map.
     * @param lastAuthTime
     *            last time user authenticated.
     * @param url
     *            URL issuing credentials.
     * @param handle
     *            handle identifying session.
     * @return formatted XML string.
     */
    public static String formatAssertion(Map attributes, String lastAuthTime, String url, String handle) {
        String xmlClientInfo = assembleAttributes(attributes);
        if (xmlClientInfo == null) {
            xmlClientInfo = "";
        }

        String issueInstant = TimeUtil.formatToGMT(new Date());

        StringBuffer assertion = new StringBuffer();
        assertion.append("<Assertion IssueInstant=\"");
        assertion.append(issueInstant);
        assertion.append("\" AuthenticationInstant=\"");
        assertion.append(lastAuthTime);
        assertion.append("\" Issuer=\"");
        assertion.append(url);
        assertion.append("\" Handle=\"");
        assertion.append(handle);
        assertion.append("\">");
        assertion.append(xmlClientInfo);
        assertion.append("</Assertion>");

        return new String(assertion);
    }

    /**
     * <p>
     * Retrieves the user attributes and formats them into an XML
     * &lt;AttributeStatement&gt; element.
     * </p>
     * 
     * @param request
     *            The request containing the attributes.
     * 
     * @return String User attributes in XML format.
     */
    private static String assembleAttributes(Map attributes) {
        log.debug("Retrieving user attributes");

        StringBuffer clientInfo = new StringBuffer();
        clientInfo.append(AttributeFormatter.XML_STMT_B_TAG);
        Iterator a = attributes.keySet().iterator();
        while (a.hasNext()) {
            String attributeName = (String) a.next();
            clientInfo.append("<");
            clientInfo.append(AttributeFormatter.XML_ATTR_TAG);
            clientInfo.append(" ");
            clientInfo.append(AttributeFormatter.XML_ATTRNAME_ATTR);
            clientInfo.append("=\"");
            clientInfo.append(xmlEncode(attributeName));
            clientInfo.append("\">");

            Iterator v = ((Set) attributes.get(attributeName)).iterator();
            while (v.hasNext()) {
                String value = (String) v.next();

                clientInfo.append(AttributeFormatter.XML_ATTRVAL_B_TAG);
                clientInfo.append(xmlEncode(value));
                clientInfo.append(AttributeFormatter.XML_ATTRVAL_E_TAG);
            }
            clientInfo.append(AttributeFormatter.XML_ATTR_E_TAG);
        }
        clientInfo.append(AttributeFormatter.XML_STMT_E_TAG);

        return clientInfo.toString();
    }

    /**
     * XML encodes the given string.
     * 
     * @param input
     *            Input string
     * 
     * @return XML encoded string.
     */
    private static String xmlEncode(String input) {
        if (input == null || input.length() == 0) {
            return input;
        }

        char[] aInput = input.toCharArray();
        StringBuffer result = new StringBuffer();
        for (int i = 0; i < aInput.length; i++) {
            if (aInput[i] < ' ' || aInput[i] > '~') {
                result.append("&#" + (byte) aInput[i] + ";");
            } else if (aInput[i] == '&') {
                result.append("&amp;");
            } else if (aInput[i] == '<') {
                result.append("&lt;");
            } else if (aInput[i] == '>') {
                result.append("&gt;");
            } else if (aInput[i] == '\'') {
                result.append("&apos;");
            } else if (aInput[i] == '"') {
                result.append("&quot;");
            } else {
                result.append(aInput[i]);
            }
        }

        return result.toString();
    }
}
